import json
# Expect: list of json objects.
import re
import typing

class MapperLocation:
    def __init__(self, location, column_name, coalesce_value=None, if_missing='ignore', attempt_json_serializing=True, delimiter='::',  convert_function=None, pass_type=None):
        """
        if_missing = str: fail | ignore | drop (default ignore)
        """
        self.location = location
        self.column_name = column_name
        self.coalesce_value=coalesce_value
        self.if_missing=if_missing
        self.attempt_json_serializing=True
        self.delimiter=delimiter
        self.convert_function=convert_function
        self.pass_type = pass_type
        
class RequiredElementMissingError(AttributeError):
    pass

class DictMapper:
    def __init__(self, in_dict_list: list, mapper: dict) -> None:
        """Creates a DictMapper instance. Use get_records() to get mapped records.

        Args:
         in_dict_list: A list of dicts to iterate over

         mapper: A dict of MapperLocation() objects defining how an exported record should look like. dict name should be the 'table_name'
            eg: {'output_table': [MapperLocation().....]}
        """
        self.in_dict_list = [d for d in in_dict_list]
        self.mapper = mapper
        self.table_names = [key for key in mapper]
        self.total_records = len(self.in_dict_list)

        self.__re_check_list_base_location = r"^.+(?=\[\])"
        self.__re_check_list_specific_element = r"[a-zA-Z]+\[[0-9]+\]"
        self.__re_check_list_unlimited_element = r"[a-zA-Z]+(?=\[\])"
        self.__re_list_location_name = r"[a-zA-Z]+(?=\[)"
        self.__re_list_location_index = r"(?<=\[)[0-9]+(?=\])"


    def get_records(self, table_name: str=None, extra_data=None) -> list:
        """ Gets the transformed flattened records.

        Args:
         table_name: 

         extra_data:

         Returns:
          A list of records that were flattened from the original json data.
        
        """
        if extra_data is None:
            extra_data = {}

        if table_name in self.table_names:
            return self._dump_to_records(in_rows=self.in_dict_list, mapper_set=self.mapper.get(table_name), extra_data=extra_data)
        else:
            raise ValueError(f'{table_name} is not defined in mapper. We got {self.table_names}')
        
    def get_column_types(self, table_name: str=None) -> dict:

        if table_name in self.table_names:
            mapper_list = self.mapper.get(table_name)
            
            mapper_return = {}
            for mapper in mapper_list:
                if mapper.pass_type is not None:
                    mapper_return.update({mapper.column_name: mapper.pass_type})

            return mapper_return
        else:
            raise ValueError(f'{table_name} is not defined in mapper. We got {self.table_names}')

    def auto_flatten_records(self, prefix: str=str(), stringify_list=False, force_lowercase=False) -> typing.Generator:
        for in_dict in self.in_dict_list:
            yield self._flatten_single_dict(root_dict=in_dict, prefix=prefix, stringify_list=stringify_list, force_lowercase=force_lowercase)

    def _flatten_single_dict(self, root_dict: dict, prefix: str=str(), stringify_list=False, force_lowercase=False) -> dict:
        """Takes a nested dictionary and "flattens" it into a single level key:value set. Object children in elements will have a prefix

        Eg: flatten_dict({"a": {"b": {"c": 1}}) == {"a_b_c": 1}
        
        root_dict: input dict

        prefix: str = name to append to returned elements    

        stringify_list: bool = False = If we encounter a list object, do we stringify it or leave it be?

        force_lowercase: bool = False = Force keys to lowercase (including prefix)

        Returns:
        dict = flattened object
        """

        root_dict = root_dict.copy()
        flattened = {}

        for k, v in root_dict.items():
            new_key = f"{prefix}{k}"
            new_key = new_key.lower() if force_lowercase else new_key
            
            # Take in any types that are not nested dictionaries.
            if type(v) not in [dict, list]:
                flattened.update({new_key: v})
            elif type(v) == dict:
                flattened.update(self._flatten_single_dict(root_dict=v, prefix=new_key + '_', force_lowercase=force_lowercase))
            # Only list should exist here.
            elif type(v) == list:
                # warnings.warn(f'{new_key} is a list')
                if stringify_list == True:
                    flattened.update({new_key: json.dumps(v)})
                else:
                    flattened.update({new_key: v})

        return flattened
    
    def _expand_dict(self, in_dict):
        flat_list = []
        row_was_expanded = False
        for k, v in in_dict.items():
            if type(v) == list:
                for row in v:
                    subroot = in_dict.copy()
                    subdata = self._flatten_single_dict(row, prefix=f'{k}_')
                    subroot.update(subdata)
                    # in_dict = subroot.copy()
                    del subroot[k]
                    flat_list.append(subroot)
                    row_was_expanded = True

        if not row_was_expanded:
            flat_list.append(in_dict)

    def _dump_single_row(self, in_row: list, mapper_set: list, extra_data: dict=None, extra_data_in_front:bool=False, ignore_location_str: str=None):
        extra_data = {} if extra_data is None else extra_data
        
        output_row = {}
        
        if len(mapper_set) == 0:
            # In the event we only are expecting to loop through subrows
            return output_row
        
        # Supposed to be a parent attribute eg {"id": "foo", "tags": ["tag_name": "bar"]}
        if extra_data_in_front:
            output_row.update(extra_data)
        
        first_column = next(iter(mapper_set)).column_name

        # for column, location in column_set.items():
        for mapperlocation in mapper_set:
            column = mapperlocation.column_name
            location = mapperlocation.location
            delimiter = mapperlocation.delimiter
            convert_function = mapperlocation.convert_function
            coalesce_value = mapperlocation.coalesce_value

            if convert_function is not None:
                coalesce_value = convert_function(coalesce_value)

            # Hack to remove unneeded location string as we now have no reference to it.
            if ignore_location_str is not None:
                # ignore_location_str = ignore_location_str #+ f'[]{mapperlocation.delimiter}'
                location = location.replace(ignore_location_str, '')

            # If we only need to pull a single element.
            try:
                location_elements = location.split(delimiter)

                if location_elements[0] == 'root' and len(location_elements) == 1:
                    if convert_function is not None:
                        in_row = convert_function(in_row)

                    output_row.update({column: in_row})

                current_object = in_row
                
                for single_location in location_elements:
                    # Ignore the root element
                    if single_location == 'root':
                        continue

                    # If we specify we want just the first element in a list eg: payments[0].get('key')
                    if re.match(self.__re_check_list_specific_element, single_location):
                        location_name = re.search(self.__re_list_location_name, single_location).group()
                        location_index = int(re.search(self.__re_list_location_index, single_location).group())
                        try:
                            current_object = current_object.get(location_name)[location_index]
                        except IndexError:
                            # print(f'Warning: indexerror {location_name}[{location_index}] does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                            if mapperlocation.if_missing == 'fail':
                                raise RequiredElementMissingError(f'Required element {location_name}[{location_index}] does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                            if mapperlocation.if_missing == 'drop':
                                # print('dropping row')
                                return None
                            else:
                                output_row.update({column: mapperlocation.coalesce_value})
                                break
                        
                    else:
                        if not hasattr(current_object, 'get') and isinstance(current_object, str) and mapperlocation.attempt_json_serializing:
                            try:
                                current_object = json.loads(current_object)
                            except json.decoder.JSONDecodeError:
                                if isinstance(current_object, str):
                                    output_row.update({column: current_object})
                                    break
                            
                        current_object = current_object.get(single_location)
                        if current_object is None:
                            # print(f'Warning: elementnotfound {single_location} does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                            if mapperlocation.if_missing == 'fail':
                                raise RequiredElementMissingError(f'Required element {single_location} does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                            elif mapperlocation.if_missing == 'drop':
                                # print(f'dropping row')
                                return None
                            else:
                                output_row.update({column: mapperlocation.coalesce_value})
                                break
                        
                    output_row.update({column: current_object})
                
                if convert_function is not None:
                    output_row[column] = convert_function(output_row[column])

            except AttributeError:
                # print(f'Warning: attributeerror {single_location} does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                if mapperlocation.if_missing == 'fail':
                    raise RequiredElementMissingError(f'Required element {single_location} does not exist in {location}. First column is {first_column}, data is {output_row.get(first_column)}')
                
                elif mapperlocation.if_missing == 'drop':
                    # print('dropping row')
                    return None

                else:
                    output_row.update({column: mapperlocation.coalesce_value})
                    continue

        if not extra_data_in_front:
            output_row.update(extra_data)
        
        return output_row

    def _dump_to_records(self, in_rows, mapper_set, extra_data=None):
        if extra_data == None:
            extra_data = {}

        out_rows = []
        
        nested_list_mapper_set = []
        nested_list_base_location = set()

        single_mapper_set = []

        for mapper in mapper_set:
            if re.search(self.__re_check_list_unlimited_element, mapper.location):
                nested_list_mapper_set.append(mapper)
                nested_list_base_location.add(re.match(self.__re_check_list_base_location, mapper.location).group())
                nested_list_base_location_delimiter = mapper.delimiter
            else:
                single_mapper_set.append(mapper)

        if len(nested_list_base_location) > 1:
            raise AttributeError(f"This library does not currently supported multiple listed mappers in a single instance. Sorry. We found {nested_list_base_location}")

        if len(nested_list_base_location) > 0:
            nested_list_base_location = next(iter(nested_list_base_location))
            nested_list_base_location_elements = nested_list_base_location.split(nested_list_base_location_delimiter)
            nested_list_base_location = nested_list_base_location + f'[]{nested_list_base_location_delimiter}'

        # feeds in the raw JSON row, and outputs a flattened record
        for row in in_rows:
            output_row = self._dump_single_row(in_row=row, mapper_set=single_mapper_set, extra_data=extra_data)

            if len(nested_list_base_location) > 0:
                output_nested_rows = []
                nested_rows = row.copy()
                for element in nested_list_base_location_elements:
                    if element == 'root':
                        continue
                    else:
                        nested_rows = nested_rows.get(element, {})
                        nested_rows = nested_rows if nested_rows is not None else {}

                if len(nested_rows) == 0:
                    nested_row = self._dump_single_row(in_row={}, mapper_set=nested_list_mapper_set, extra_data=output_row, extra_data_in_front=True, ignore_location_str=nested_list_base_location)
                    if nested_row is None:
                        continue
                    
                    output_nested_rows.append(nested_row)

                else:
                    for nested_row in nested_rows:
                        if nested_row is None:
                            continue
                        output_nested_rows.append(self._dump_single_row(in_row=nested_row, mapper_set=nested_list_mapper_set, extra_data=output_row, extra_data_in_front=True, ignore_location_str=nested_list_base_location))
                
                for r in output_nested_rows:
                    # These are bad hacks, but i want to go home.
                    if r is None:
                        continue
                    out_rows.append(r)

                if len(output_nested_rows) == 0 and output_row is not None and len(output_row) != 0:
                    out_rows.append(output_row)

            else:
                if output_row is not None and len(output_row) != 0:
                    out_rows.append(output_row)

        return out_rows


import datetime
import isoweek
import pytz

''' weeks start on mondays '''
def first_date_of_week(any_datetime):
  year, iso_week, iso_weekday = any_datetime.isocalendar()
  return any_datetime - datetime.timedelta(days=(iso_weekday-1))

def last_date_of_week(any_datetime):
  year, iso_week, iso_weekday = any_datetime.isocalendar()
  return first_date_of_week(any_datetime) + datetime.timedelta(days=6) # last date is 6 days after the first date

def is_leap_day(date):
  return date.month == 2 and date.day == 29

def is_in_leap_week(date):
  return isoweek.Week.withdate(date).week == 53

def is_in_week_prior_leap_week(date):
  same_day_next_week = date + datetime.timedelta(days=7)
  return is_in_leap_week(same_day_next_week)

def has_leap_week(year: int):
  return isoweek.Week(year, 53).week == 53

def last_year(date):
  return isoweek.Week.withdate(date).year - 1

def days_since_same_date_last_year(date):
  return 7 * 53 if is_in_leap_week(date) or has_leap_week(last_year(date)) else 7 * 52

def comp_yearwise(datetime_or_date, n_years):
    dt : datetime.datetime = safe_datetime(datetime_or_date)
    comp_days = 0
    dt_week = isoweek.Week.withdate(dt)
    while n_years >= 1:
        if n_years==1 and comp_days == 0:
            weeks = 52
        else:
            weeks = isoweek.Week.last_week_of_year(dt_week.year - n_years).week
        comp_days += weeks*7
        n_years = n_years-1
    comp_date = dt - datetime.timedelta(days=comp_days)
    return comp_date

def days_between(datetime_or_date_a, datetime_or_date_b):

    dt_a : datetime.datetime = safe_datetime(datetime_or_date_a)
    dt_b : datetime.datetime = safe_datetime(datetime_or_date_b)

    delta = dt_a - dt_b
    
    return delta.days

def safe_datetime(datetime_or_date):
    if type(datetime_or_date) == datetime.date:
        return datetime.datetime.combine(datetime_or_date, datetime.datetime.min.time())
    elif type(datetime_or_date) == datetime.datetime:
        return datetime_or_date
    else:
        raise TypeError
    
def get_l4w_days(today: datetime.date):
    """Allows for only showing MTD sales where there are not enough days to show L4W
    
    Specifically we do not want to look in 2022 when showing 2023 W1-W4 Sales. Assumes
    dates are inclusive.

    Params:
        today (datetime.date): date to consider

    Returns:
        l4w_days (int): how many days to go back.
    """
    if today.month == 1 and today.day < 28:
        return today.day - 1
    else:
        return 27

def timestamp_to_datetimetz(timestamp: float, to_tzinfo: str='UTC', timezone_aware=False) -> datetime.datetime:
    """Convert a timestamp into a datetime object with a specific timezone applied.
    
    Params:

        timestamp (float): Timestamp in seconds

        to_tzinfo (str): To timezone in tzinfo eg America/New_York. Default/None is UTC.

        timezone_aware (bool): Include timezone info in returned datetime object

    Returns:
        A datetime object of the converted date. 
    """
    
    from_tz = pytz.UTC
    to_tz = pytz.timezone(to_tzinfo)

    out_ts = datetime.datetime.fromtimestamp(timestamp, tz=from_tz)

    out_ts = out_ts.astimezone(to_tz)

    if timezone_aware:
        return out_ts
    
    return out_ts.replace(tzinfo=None)

def chunk_days(start_date: datetime.date, end_date: datetime.date, days: int):
    """ Returns an iterable of datetime.date objects based on a start_date, end_date and chunk days interval

    Params:
      start_date (datetime.date): Starting interval
      end_date (datetime.date): The ending interval
      days (int): The amount of days
    
    """
    timedelta_curr_days = datetime.timedelta(days=days)

    _curr_start = start_date

    if days < 1:
        raise ValueError("Days must be greater than 0.")
    
    if not all([isinstance(start_date, datetime.date), isinstance(end_date, datetime.date)]):
        raise ValueError("start_date and end_date must be of type datetime.date")

    while True:
        _curr_end = _curr_start + timedelta_curr_days

        if start_date == end_date:
           yield (start_date, end_date)
           break

        if _curr_end > end_date:
            yield (_curr_start, end_date)
            break

        if _curr_start < end_date:
            yield (_curr_start, _curr_end)
            _curr_start = _curr_start + timedelta_curr_days




from typing import Dict, TypedDict
import sendgrid
from sendgrid.helpers import mail
import re
import base64
import io
import os
import zipfile as zf
import openpyxl as xl
from utilita import excel, date_fns
import mimetypes

class SingleEmail(TypedDict):
    """Define email address and friendly name

    Params:
        email (str): test@example.com
        name (str): John Doe
        reply_to (str): support@example.com
    """
    email: str
    name: str
    reply_to: str

class RecipientsEmail(TypedDict):
    """Define the recipients for an email.

    You cannot have a single email be in multiple lines eg: john@example.com cannot be in cc
    and bcc.
    
    Params:
        to (str): test1@example.com, test2@example.com, test3@example.com
        cc (str): test4@example.com
        bcc (str): test5@example.com, test6@example.com        
        """
    to: str
    cc: str
    bcc: str

class ReplyToEmail(TypedDict):
    """Define email address and friendly name

    Params:
        email (str): support@example.com
    """
    email: str

class SendgridHelper:
    def __init__(self, sendgrid_api_key: str=None, from_email: SingleEmail=None, reply_to_email: ReplyToEmail=None):
        """ Creates a new SendgridHelper Instance

        Args:
            sendgrid_api_key (str): Sendgrid API key from the website

            from_email (dict): dict in the form of {"email": 'noreply@email.com', 'name': 'John Doe', 'reply_to': 'support@email.com'}. replyto is optional.

            reply_to_email (dict): dict in the form of {"email": 'support_group@email.com', 'name': 'John Doe'}. For setting an alternate reply address.
        """

        if not isinstance(from_email, dict) or from_email.get('email') is None:
            raise ValueError('from email must be defined.')
        
        if from_email.get('reply_to') is not None:
            reply_to_email: SingleEmail = {"email": from_email.get("reply_to")}

        if isinstance(reply_to_email, dict) and reply_to_email.get('email') is None:
            raise ValueError('reply-to email must be defined if its defined.')

        self.from_email: SingleEmail = from_email
        self.reply_to: SingleEmail = reply_to_email
        self.sendgrid_api_key = sendgrid_api_key

        assert not [x for x in (sendgrid_api_key, from_email) if x is None], f"Must have a sendgrid_api_key and from_email"

    def config_email(self, subject: str, recipients: RecipientsEmail, body: str):
        """Configure basic settings of the email.

        Args:
        subject (str): Subject line

        recipients (dict): dict of the email recipients in the form of: \
            :python:`{"to": "example1, example2", "cc": "example3", "bcc": "example4, example5"}`. \
                Each line is optional. Separate emails by commas. The same email cannot exist \
                in multiple lines.

        body (str): HTML of the body of the email.
        """


        e = """Configure basic settings of the email.

        Params:
            subject (str): Subject line

            recipients (dict): dict of the email recipients in the form of: {"to": "example1, example2", "cc": "example3", "bcc": "example4, example5"}. Each line is optional. Separate emails by commas. The same email cannot exist in multiple lines.

            body (str): HTML of the body of the email.        
        """
        self.client = sendgrid.SendGridAPIClient(api_key = self.sendgrid_api_key)
        self.msg = sendgrid.helpers.mail.Mail()
        self.msg.from_email = sendgrid.helpers.mail.Email(self.from_email.get('email'), self.from_email.get('name'))

        if self.reply_to is not None:
            self.msg.reply_to = sendgrid.helpers.mail.reply_to.ReplyTo(
                email=self.reply_to.get('email'), name=self.reply_to.get('name')
                )

        
        self.p = sendgrid.helpers.mail.Personalization()
        self.p.subject = subject
        self.recipients = recipients

        if isinstance(self.recipients, Dict):
            to_emails = get_email_address_list(self.recipients.get('to'))
            for e in to_emails:
                self.p.add_to(sendgrid.helpers.mail.Email(e))

            cc_emails = get_email_address_list(self.recipients.get('cc', []))
            for e in cc_emails:
                self.p.add_cc(sendgrid.helpers.mail.Email(e))

            bcc_emails = get_email_address_list(self.recipients.get('bcc', []))
            for e in bcc_emails:
                self.p.add_bcc(sendgrid.helpers.mail.Email(e))

        else:
            to_emails = get_email_address_list(self.recipients)
            for d in to_emails:
                self.p.add_to(sendgrid.helpers.mail.Email(d))


        self.msg.add_personalization(self.p)

        self.msg.add_content(
            sendgrid.helpers.mail.Content(
                'text/html',
                body
            )
        )

    def send_email(self):
        """Attemps to send the email"""
        try:
            response = self.client.client.mail.send.post(request_body = self.msg.get())
            return response
        except Exception() as e:
            return e


    def attach_df_as_csv(self, file_obj, df_filename=None, compressed=True):
        """Attach a pandas dataframe as a csv file

        Params:
            file_obj: dict in the format of {"df": [dataframe], "zip_filename": "[filename as string]"}
                {"df": unmapped_df, "zip_filename": f"uber_unmappable_transactions for {date}"}

            df_filename:

            compressed=True: compress the file as a zip file.
        """
        df = file_obj.get("df")
        zip_filename=file_obj.get("zip_filename")
        
        assert not [x for x in (df, zip_filename) if x is None], "df==None and zip_filename==None must not be true."
        
        zip_filename = zip_filename.replace('.zip','')

        if df_filename is None: df_filename = zip_filename
        df_filename = df_filename.replace('.csv','')

        zip_bytes = df_to_zipfile_bytes(df, df_filename+".csv")
        zip_bytes.seek(0)
        report_bytes = zip_bytes.read()

        self.msg.add_attachment(
            attachment_from_bytes(
            file_name = zip_filename+".zip",
            file_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            file_bytes = report_bytes
            )
        )


    def attach_single_df_as_csv(self, df, df_filename, compressed=False) -> None:
        """Attach a pandas dataframe as a csv file

        Params:
            df: dataframe object

            df_filename: what to name the file. 

            compressed=True: compress the file as a zip file. If True, the compressed file name is the same as the df filename.

        Returns:
            Nothing, as it gets added to the object.
        """
        
        if compressed==True:
            zip_filename = df_filename.replace('.zip','')
            df_filename = df_filename.replace('.csv','')

            zip_bytes = df_to_zipfile_bytes(df=df, df_filename=df_filename+'.csv')
            zip_bytes.seek(0)
            report_bytes = zip_bytes.read()
            
            attach_filename = df_filename+'.zip'
            attach_mime_type = 'application/zip'

        else:
            csv_bytes = df_to_csv_bytes(df=df)
            attach_filename = df_filename+'.csv'
            attach_mime_type = 'text/csv'
            report_bytes = csv_bytes.read()

        self.msg.add_attachment(
            attachment_from_bytes(
            file_name = attach_filename,
            file_type = attach_mime_type,
            file_bytes = report_bytes
            )
        )
    
    def attach_excel_file_from_path(self, excel_path):
        """Attaches an excel file from the filesystem.
        
        This is a deprecated function. Please use attach_file_from_path() instead.

        Params:
            excel_path (str): file from filesystem.
        """
        self.attach_file_from_path(excel_path)

    def attach_excel_file_from_bytes(self, excel_bytes, file_name):
        self.msg.add_attachment(
            attachment_from_bytes(
                file_name = file_name,
                file_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                file_bytes = excel_bytes
            )
        )

    def attach_openpyxl_workbook_from_bytes(self, wb: object, file_name: str):
        """Attaches an openpyxl workbook object into the email.
        
        Params:
            wb (openpyxl.workbook.workbook.Workbook): Workbook file.

            file_name (str): The actual file name to show in the email.   
        """
        report_bytes = io.BytesIO()
        wb.save(report_bytes)
        report_bytes.seek(0)
        self.attach_excel_file_from_bytes(report_bytes.read(), file_name=file_name)

    def attach_file_from_path(self, file_path: str) -> None:
        """Attaches a file from the filesystem.
        
        Params:
            file_path (str): full or relative filepath of the file
        """
        file_name = os.path.basename(p=file_path)

        with open(file=file_path, mode='rb') as f:
            data: bytes = base64.b64encode(f.read())

        mime_type, _ = mimetypes.guess_type(file_path)

        a = mail.Attachment()
        a.disposition = mail.Disposition(disposition='attachment')
        a.file_type = mail.FileType(file_type=mime_type)
        a.file_name = mail.FileName(file_name=file_name)
        a.file_content = mail.FileContent(file_content=data.decode('utf-8'))

        self.msg.add_attachment(a)


def get_email_address_list(address_string):
    if isinstance(address_string, str):
        address_string = re.split(r',|;', address_string)

    return [x.strip() for x in address_string]

def attachment_from_bytes(file_name, file_type, file_bytes):
    a = mail.Attachment()
    a.disposition = mail.Disposition('attachment')
    a.file_type = mail.FileType(file_type)
    a.file_name = mail.FileName(file_name)
    data_64 = str(base64.b64encode(file_bytes).decode('utf-8'))
    a.file_content = mail.FileContent(data_64)
    return a

def df_to_zipfile_bytes(df, df_filename):
    file_obj = io.BytesIO()
    x = zf.ZipFile(file=file_obj, mode='w', compression=zf.ZIP_DEFLATED, compresslevel=5) #winrar "normal" compression
    x.writestr(zinfo_or_arcname=df_filename ,data=df.to_csv(index=False)) #filename of the 
    x.close()

    return file_obj

def df_to_csv_bytes(df):
    file_obj = io.BytesIO(initial_bytes=df.to_csv(index=False).encode('utf-8'))
    file_obj.seek(0)

    return file_obj


from . import date_fns
from . import misc
from . import excel
from . import date_strs
from . import date_fns
from . import sendgridhelper as email
from . import sendgridhelper
from . import csvmapper
from . import dictmapper


import openpyxl as xl
import pandas as pd
import numpy as np
from deprecated import deprecated
import os
from typing import TypedDict
from io import BytesIO
from tempfile import NamedTemporaryFile
import re

class ExcelTableRows(TypedDict):
    headers: list
    rows: list

def workbook_to_bytes(wb: xl.Workbook, lock_worksheets: bool=False) -> BytesIO:
    """Takes an openpyxl workbook, and saves it as a BytesIO object.

    Params:
        wb (openpyxl.Workbook): Workbook object to save

        lock_worksheets (bool): Lock worksheets with no password.
    """
    if lock_worksheets:
        for x in wb.worksheets:
            x.protection.enable() # always do this before exporting because they automatically become unprotected somehow

    # If on windows, add delete=False to NamedTemporaryFile()
    ntf_delete = False if os.name == 'nt' else True
    with NamedTemporaryFile(delete=ntf_delete) as tmp:
        wb.save(tmp.name)
        excel_bytes = BytesIO(tmp.read())
        excel_bytes.seek(0)
        excel_bytes = excel_bytes.read()
    return excel_bytes

def df_to_worksheet(df: pd.DataFrame, cell: xl.cell.cell.Cell) -> None:
    """Takes a DataFrame and writes it to a worksheet.
    
    Params:
        df (DataFrame): DataFrame to insert to worksheet

        cell (openpyxl.cell.cell.Cell): Cell to begin inserting data at.
    """
    for r in range(df.shape[0]):
        for c in range(df.shape[1]):
            cell.offset(r+1, c).value = df.iloc[r, c]

@deprecated(version='0.3.0', reason='renamed to df_to_worksheet')
def df_to_sh(df: pd.DataFrame, cell: xl.cell.cell.Cell):
    df_to_worksheet(df, cell)

def df_to_excel_table_resize(df: pd.DataFrame, wb: xl.Workbook, sheet_name: str, table_name: str, convert_np_bools: str=True) -> None:
    """Takes a pandas dataframe, and writes it into an excel table.
    
    It will also resize the table rowcount to match. It will not resize the columns. Pandas will also not
    respect the excel column arrangement, it will insert the data in the dataframe's column order.

    Params:
        df (DataFrame): Dataframe to insert

        wb (openpyxl.Workbook): Opened workbook object

        sheet_name (str): Worksheet name in workbook

        table_name (str): Table name in Worksheet

        convert_np_bools (bool): Booleans in dataframes get converted to an int. This is to force them as python bools.
            WARNING: This will overwrite the original dataframe

    Returns:
        None. TODO: Return workbook object?

    Exceptions:
        ValueError: If a dataframe's column count is different from the excel table's column count.
    
    """
    data_db = wb[sheet_name]
    assert table_name in data_db.tables, f"Table {table_name} does not exist in the worksheet {sheet_name}"
    ref_start, ref_end = data_db.tables[table_name].ref.split(':')

    start_col, start_row = split_string(ref_start)
    end_col, end_row = split_string(ref_end)
    table_length = (excel_letter_to_numeric(end_col))-(excel_letter_to_numeric(start_col)-1)

    if table_length != len(df.columns):
        raise ValueError(f"Dataframe width does not match table: df has {len(df.columns)} columns and {table_name} has {table_length} columns.")
    
    if convert_np_bools:
        for col in df.columns:
            if df[col].dtype == np.bool_:
                df[col] = df[col].astype(object)

    if len(df) > 0:
        # Blank out range (if data already exists)
        # TODO: Add check for keeping headers?
        for row in data_db[f"{start_col}{start_row+1}:{end_col}{end_row}"]:
            for cell in row:
                cell.value = None

        df_to_worksheet(df, data_db[f"{start_col}{start_row}"])
        data_db.tables[table_name].ref = f"{start_col}{start_row}:{end_col}{start_row+len(df)}"

def excel_table_to_rows(wb: xl.Workbook, sheet_name: str, table_name: str) -> ExcelTableRows:
    """Takes an openpyxl workbook, sheet_name, and table_name, and returns a dict of headers and rows.

    Args:
        wb (workbook): openpyxl workbook

        sheet_name (str): worksheet name

        table_name (str): table name

    Returns:
        dict of {'headers': <list of headers>, 'rows': <list of rows>}
    """
    if sheet_name not in wb.sheetnames:
        raise ValueError(f'Workbook does not have the sheet name {sheet_name}. worksheet names are case sensitive.')

    if sheet_name in wb.sheetnames and table_name not in wb[sheet_name].tables:
        raise ValueError(f'Worksheet does not have the table name {table_name}. table names are case sensitive.')
    
    sheet = wb[sheet_name]
    tbl_ref = sheet.tables[table_name].ref
    start_ref, end_ref = map(excel_cell_to_coordinates, tbl_ref.split(':'))

    headers = None
    data = []

    for row in sheet.iter_rows(min_row=start_ref[1], max_row=end_ref[1], min_col=start_ref[0], max_col=end_ref[0]):
        if not isinstance(headers, list):
            headers = [cell.value for cell in row]
        else:
            data.append((cell.value for cell in row))
    
    return {
        'headers': headers,
        'rows': data
    }


def excel_table_to_df(wb: xl.Workbook, sheet_name: str, table_name: str) -> pd.DataFrame:
    """Takes an openpyxl workbook, sheet_name, and table_name, and returns a dataframe

    Args:
        wb (workbook): openpyxl workbook

        sheet_name (str): worksheet name

        table_name (str): table name

    Returns:
    """
    row_data = excel_table_to_rows(wb=wb, sheet_name=sheet_name, table_name=table_name)

    return pd.DataFrame(data=row_data['rows'], columns=row_data['headers'])


def get_named_range_cell(wb: xl.Workbook, rng_name: str) -> xl.cell.cell.Cell:
    '''return cell object from range name'''
    r = list(wb.defined_names[rng_name].destinations)[0]
    return wb[r[0]][r[1]]


def excel_letter_to_numeric(letters) -> int:
    """Takes a full excel letter column refernece and convert it into a number.
    
    Params:
        letters (str): Excel column in letters eg: a=1, aa=26, aaa=703
        
    Returns:
        Converted int of the individual letters.
    """
    result = 0
    for letter in letters:
        result = result * 26 + __letter_to_number(letter)
    return result

def numeric_to_excel_letter(number: int) -> str:
    """Takes an column number, and returns a letter that matches excel, where 1=A.
    
    Params:
        number (int): Column Number
        
    Returns:
        Letters that can be used in excel where 1=A, 27=AA etc...
    """
    if isinstance(number, int) and number < 1:
        raise ValueError('Number must be positive')
    
    result = ''
    while number > 0:
        number, remainder = divmod(number - 1, 26)
        result = __number_to_letter(remainder + 1) + result
    return result

def excel_cell_to_coordinates(cell: str) -> tuple:
    """Takes an excel reference eg: A1, and returns a tuple with numeric coordinates eg: (1, 1).
    
    Params:
        cell (str): Cell String

    Returns:
        tuple of cell coordinates.
    """
    if re.match(r'^([A-Za-z])+(\d+)$', cell) is not None:
        match = re.split('(\d+)', cell)
        return (excel_letter_to_numeric(match[0]), int(match[1]))
    
    else:
        raise ValueError('Cell reference must be in excel format eg: A1, ZZ21')

def split_string(string: str) -> list:
    """Converts excel cell to list of col, row eg: B2 -> ['B', 2]"""
    col, row = re.findall(r'\D+|\d+', string)
    return [col, int(row)]

def __letter_to_number(letter: str) -> int:
    """Takes an excel single cell letter, and converts it into a number.
    
    Params:
        letter (str): single letter A-Z

    Returns:
        letter position, where A=1 and Z=26.
    """
    return ord(letter.upper()) - ord('A') + 1

def __number_to_letter(number: int) -> str:
    """Takes a numeric column and converts it into an letter.
    
    Params:
        number (str): 1-26 to convert to a letter.

    Returns:
        Converted letter from number, where 1=A, 26=Z.

    Exceptions:
        ValueError: If the input number is not between 1 and 26.
    """
    if isinstance(number, int) and number >= 1 and number <= 26:
        return chr(number + ord('A') - 1)
    else:
        raise ValueError('number must be 1-26')


import sys
from os import path
from typing import Type
sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )

import unittest

import datetime
from utilita import date_fns

class TestEverything(unittest.TestCase):

  def test_base(self):

    y2020w52d7 = datetime.date(2020,12,27)
    y2020w53d1 = datetime.date(2020,12,28)

    self.assertFalse(date_fns.is_in_leap_week(y2020w52d7)) # => False
    self.assertTrue(date_fns.is_in_leap_week(y2020w53d1)) # => True

    self.assertTrue(date_fns.is_in_week_prior_leap_week(y2020w52d7)) # => True
    self.assertFalse(date_fns.is_in_week_prior_leap_week(y2020w53d1)) # => False


    self.assertEqual(date_fns.days_since_same_date_last_year(y2020w52d7), 364) # => 364 (days in non-leep-week years)
    
    y2020w53 = [y2020w53d1 + datetime.timedelta(days=x) for x in range(7)]

    for date in y2020w53:
      self.assertEqual(date_fns.days_since_same_date_last_year(date), 371) # => 371 (days in leep-week years)


    y2021w1d1 = datetime.date(2021,1,4)
    y2021w52d7 = datetime.date(2022,1,2)
    y2022w1d1 = datetime.date(2022,1,3)

    self.assertFalse(date_fns.is_in_leap_week(y2021w1d1)) # => False
    self.assertEqual(date_fns.days_since_same_date_last_year(y2021w1d1), 371)
    self.assertEqual(date_fns.days_since_same_date_last_year(y2021w52d7), 371)
    self.assertEqual(date_fns.days_since_same_date_last_year(y2022w1d1), 364)


  def test__comp_yearwise__non_leapyear_non_leapweek_1y(self):
    run_dt = datetime.datetime(2019, 6, 6)
    expected_dt = datetime.datetime(2018, 6, 7)

    comp_dt = date_fns.comp_yearwise(run_dt, n_years = 1)

    self.assertEqual(comp_dt, expected_dt, f"expected {expected_dt}, got {comp_dt} instead")


  def test__comp_days_yearwise__non_leapyear_non_leapweek_2y(self):
    run_date = datetime.datetime(2019, 6, 6)
    expected_dt = datetime.datetime(2017, 6, 8)

    comp_dt = date_fns.comp_yearwise(run_date, n_years = 2)

    self.assertEqual(comp_dt, expected_dt, f"expected {expected_dt}, got {comp_dt} instead")


  def test__comp_days_yearwise__into_leapyear_from_normal(self):
    run_date = datetime.datetime(2021, 6, 6) # 2021 is after 2020 (leap week year!)
    expected_date = datetime.datetime(2020,6,7)

    comp_dt = date_fns.comp_yearwise(run_date, n_years = 1)

    self.assertEqual(comp_dt, expected_date, f"Leap year not handled: expected {expected_date}, got {comp_dt} instead")
      
  def test__days_between(self):
    a = datetime.date.today() 
    b = datetime.date.today() - datetime.timedelta(days=1)
    days = date_fns.days_between(a, b)
    expected_days = 1

    self.assertEqual(days, expected_days, f"expected days between {a} and {b} to be {expected_days} not {days}")

    days_flip = date_fns.days_between(b, a)
    expected_days_flip = -1

    self.assertEqual(days_flip, expected_days_flip, f"expected days between {b} and {a} to be {expected_days_flip} not {days_flip}")

  def test__safe_datetime(self):
    dt = datetime.datetime(2021, 1, 1)
    safe_dt = date_fns.safe_datetime(dt)
    self.assertEqual(type(safe_dt), datetime.datetime, f"{safe_dt} is not a datetime object")

    d = datetime.date(2021, 1, 1)
    safe_dt2 = date_fns.safe_datetime(d)
    self.assertEqual(type(safe_dt2), datetime.datetime, f"{safe_dt2} is not a datetime object")

    num = 1 
    self.assertRaises(TypeError, date_fns.safe_datetime, num)

unittest.main()


import typing
import csv
import re
import copy

class MapperLocation:
    def __init__(self, csv_column_name: str=None, csv_column_position: int=None, column_name: str=None, coalesce_value: typing.Any=None, 
                 if_missing: str='ignore', if_column_missing: str='fail', convert_function: typing.Callable=None, pass_type=None, 
                 composite_element_name: str=None):
        
        """Maps a specific location from a dict object to a DictCruncher "column"

        Args:
        csv_column_name (str): A specific column that already exists in the CSV file. Either this or csv_column_position must be specified.

        csv_column_position (int): A specific column position in a csv file. Either this or csv_column_name must be specified.

        column_name (str): a string column name to assign to a mapping.

        coalesce_value (Any): Default value if location is missing. Can be any type.
        
        if_missing (str): fail | ignore (default ignore). What to do if a mapping is missing.

        if_column_missing (str): fail | ignore (default fail). What to do if a named column is missing.

        delimiter (str): Delimiter to pass to csv library.

        convert_function (Callable): Function to apply on this value. This will also apply to a
            coalesce_value value.

        pass_type (Any): Value to return when using DictCruncher.get_column_types()

        composite_element_name (str): return a value based off of a single or multiple elements. use curly braces eg: {trn_pk}-{trn_cpk}

        Returns:
            MapperLocation object. This should be used as a 'table definition' eg:

            [
                MapperLocation(csv_column_name='CSV Column', column='id_number')
            ]
        """

        if column_name is None:
            raise ValueError('column_name must be defined.')

        if sum(1 for v in [csv_column_name, csv_column_position, composite_element_name] if v is not None) > 1:
            raise ValueError('%s: Either csv_column_position or csv_column_name must be used. They both cannot be used at the same time.' % column_name)

        self.csv_column_name = csv_column_name
        self.csv_column_position = csv_column_position
        self.column_name = column_name
        self.coalesce_value = coalesce_value
        self.if_missing = if_missing
        self.if_column_missing = if_column_missing
        self.convert_function = convert_function
        self.pass_type = pass_type
        self.composite_element_name = composite_element_name
        
        self.header_row = None

    def __str__(self):
        ident = f'"{self.csv_column_name}"' if self.csv_column_name is not None else f'csv_position: {self.csv_column_position}'
        return f"MapperLocation {ident} -> {self.column_name}"

    def _register_column_position(self, position: typing.Union[int, list]):
        """Assigns a column positition to a MapperLocation mapper object.

        Args:
        
        position (int | dict): The column position starting from 0 to get data from.

        TODO: This feels really janky, but think its better to have the 
            column position exist within the MapperLocation object for now.
        """

        self.csv_column_position = position
        
class RequiredElementMissingError(AttributeError):
    """Raised when a MapperLocation() value is missing AND if_missing is True    
    """
    pass

class CSVMapper:
    def __init__(self, filepath_or_buffer, mapper: list, has_headers=True, csv_delimiter: str=None, csv_quotechar: str=None, csv_newline=None, filepath_encoding='utf-8-sig'):
        """A library for cleaning up CSV columns without much heartache

        Args:
         filepath_or_buffer: Either a filename/path or a file handle

         mapper: A list of MapperLocation() objects defining how an exported record should look like. The order of elements in the list
            will be how they are exported.

         has_headers: Whether the csv has headers

         csv_delimiter (str): Delimiter to use on the file.

         csv_quotechar (str): Quote Character to use on file

         csv_newline (str): Newline to pass into csv module

         filepath_encoding (str): If passing in a filepath, the encoding (eg: utf-8). default is utf-8-sig
        """

        if isinstance(filepath_or_buffer, str):
            self.__in_file = open(filepath_or_buffer, 'r', encoding=filepath_encoding)

        elif hasattr(filepath_or_buffer, 'read'):
            self.__in_file = filepath_or_buffer

        else:
            raise IOError('Could not open file for reading.')

        csvfile_params = {
            'delimiter': csv_delimiter,
            'quotechar': csv_quotechar,
            'newline': csv_newline,
        }

        self.__csv_filtered_params = {k: v for k, v in csvfile_params.items() if v is not None}

        self.mapper = mapper

        self.has_headers = has_headers
        
        self._map_to_file()

    def _get_csvfile(self):
        """Get a csv.reader handler"""
        self.__in_file.seek(0)
        return csv.reader(self.__in_file, quoting = csv.QUOTE_ALL, **self.__csv_filtered_params)

    def _map_to_file(self):
        if self.has_headers:
            csvfile = self._get_csvfile()
            self.header_row = next(csvfile)
            
            for singlemapper in self.mapper:
                if singlemapper.composite_element_name is not None:
                    try:
                        mapped_position = {}
                        composite_elements = re.findall(r'\{(.*?)\}', singlemapper.composite_element_name)
                        
                        for element in composite_elements:
                            mapped_position.update({element: self.header_row.index(element)})

                        singlemapper._register_column_position(mapped_position)

                    except ValueError:
                        if singlemapper.if_column_missing == 'ignore':
                            continue

                        raise ValueError('Unable to find the column "%s" in the input file. Found columns were %s' % (singlemapper.csv_column_name, str(self.header_row)))

                if singlemapper.csv_column_name is not None:
                    try:
                        mapped_position = self.header_row.index(singlemapper.csv_column_name)
                        singlemapper._register_column_position(mapped_position)

                    except ValueError:
                        if singlemapper.if_column_missing == 'ignore':
                            continue

                        raise ValueError('Unable to find the column "%s" in the input file. Found columns were %s' % (singlemapper.csv_column_name, str(self.header_row)))
        else:
            for singlemapper in self.mapper:
                if singlemapper.csv_column_name is not None:
                    raise ValueError('%s has a CSV header defined, but CSVMapper has has_header=False' % singlemapper)

    def to_records(self) -> typing.Generator:
        """Returns an iterable where each dict contains a row. Iterable is a generator
        
        Args:
            (Nothing)

        Returns:
            Iterable over each row

        Raises:
            RequriedElementMissingError - If MapperLocation's if_missing parameter is set to fail.
        """
        csvfile = self._get_csvfile()
        if self.has_headers:
            next(csvfile)

        for row in csvfile:
            out_row = {}
            
            for singlemapper in self.mapper:
                return_value = None

                # this is a dict if using composite_element_name
                if isinstance(singlemapper.csv_column_position, dict):
                     # this is a deepcopy because python only copies a reference, not the full object.
                    element_values = copy.deepcopy(singlemapper.csv_column_position)
                    for element in element_values.keys():
                        # The idea is csv_colum_position will be {'column_name': (int position in list)}. The below will overwrite the position with the row value.
                        element_values[element] = row[element_values[element]]

                    return_value = singlemapper.composite_element_name.format(**element_values)

                # if singlemapper.csv_column_position is not None:
                if isinstance(singlemapper.csv_column_position, int):
                    return_value = row[singlemapper.csv_column_position]
                
                return_value = None if return_value == '' else return_value                    

                if return_value is None:
                    if singlemapper.if_missing == 'fail':
                        raise RequiredElementMissingError('Value is missing for column %s. row contains %s' % (singlemapper.coalesce_value, out_row))
                    elif singlemapper.coalesce_value is not None:
                        return_value = singlemapper.coalesce_value

                if singlemapper.convert_function is not None:
                    return_value = singlemapper.convert_function(return_value)

                out_row.update({singlemapper.column_name: return_value})
            yield out_row

    def to_dataframe(self):
        """Returns a pandas DataFrame with the converted data.
        
        Args:
            (Nothing)

        Returns:
            Pandas DataFrame

        Raises:
            ImportError: If pandas is not installed.
        """
        try:
            import pandas as pd
            return pd.DataFrame.from_records(self.to_records())

        except ImportError:
            raise ImportError('Must have pandas installed in order to use this function.')
        


import datetime
import sys
import argparse

from os import path
sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )

import unittest

import datetime
import time
from utilita import misc

import pendulum

class TestEverything(unittest.TestCase):

  def test__ymd_arg(self):

    self.assertEqual(misc.ymd_arg('2021-01-11'), datetime.date(2021,1,11))
    self.assertEqual(misc.ymd_arg('2021-1-11'), datetime.date(2021,1,11))

    with self.assertRaises(argparse.ArgumentTypeError):
      misc.ymd_arg('1-11-2021')

    with self.assertRaises(argparse.ArgumentTypeError):
      misc.ymd_arg('2021/01/11')

    with self.assertRaises(argparse.ArgumentTypeError):
      misc.ymd_arg('2021.01.11')

    self.assertEqual(misc.ymd_arg(None), None)

  def test__parse_templated(self):

    templated_greetings = {
      'keywords': {
        'thing': 'world'
      },
      'templates': {
        'hi': 'hello {thing}',
        'bye': 'laters {thing}'
      }
    }

    parsed_greetings = misc.parse_templated(templated_greetings)

    expected_greetings = {
      'hi': 'hello world',
      'bye': 'laters world'
    }

    self.assertEqual(parsed_greetings, expected_greetings)

    with self.assertRaises(TypeError):
      misc.parse_templated(None)

    with self.assertRaises(AssertionError):
      misc.parse_templated({})

    templated_dates = {
      'keywords': {
        'year': '2021'
      },
      'templates': {
        'start_date': '{year}-01-01',
        'end_date': '{year}-02-01'
      }
    }

    parsed_dates = misc.parse_templated(templated_dates, map_value=misc.ymd_arg)

    expected_dates = {
      'start_date': datetime.date(2021,1,1),
      'end_date': datetime.date(2021,2,1)
    }

    self.assertEqual(parsed_dates, expected_dates)


    # old name should yield a deprecation warning
    templated_compat = {
      'template_dict': { # old name
        'year': '2021'
      },
      'templates': {
        'start_date': '{year}-01-01',
      }
    }

    with self.assertWarns(PendingDeprecationWarning):
      misc.parse_templated(templated_compat, map_value=misc.ymd_arg)

  def test__parse_templated__w_formatted_keywords(self):

    templated_schematable_urls = {
      'keywords': {
        'dw':'postgresql://{user}:{pass}@localhost:5432/dw',
        'user': 'sysadmin', 
        'pass': 'password123'
      },
      'templates': {
        'product': '{dw}#product',
        'producer': '{dw}#producer',
      }
    }

    parsed_schematable_urls = misc.parse_templated(templated_schematable_urls)

    expected_schematable_urls = {
      'product': 'postgresql://sysadmin:password123@localhost:5432/dw#product',
      'producer': 'postgresql://sysadmin:password123@localhost:5432/dw#producer',
    }

    self.assertEqual(parsed_schematable_urls, expected_schematable_urls)

  def test__parse_templated__w_deep_recursive_formatted_keywords(self):

    templated_schematable_urls = {
      'keywords': {
        'dw_schema_url': '{dw_db_url}/dw',
        'dw_db_url':'postgresql://{user}:{pass}@localhost:5432',
        'user': 'sysadmin', 
        'pass': 'password123'
      },
      'templates': {
        'product': '{dw_schema_url}#product',
        'producer': '{dw_schema_url}#producer',
      }
    }

    parsed_schematable_urls = misc.parse_templated(templated_schematable_urls)

    expected_schematable_urls = {
      'product': 'postgresql://sysadmin:password123@localhost:5432/dw#product',
      'producer': 'postgresql://sysadmin:password123@localhost:5432/dw#producer',
    }

    self.assertEqual(parsed_schematable_urls, expected_schematable_urls)

  def test__parse_templated__w_extra_keywords(self):

    templated_db_urls = {
      'keywords': {
        'port':'5432'
      },
      'templates': {
        'dw_db_url': 'postgresql://{user}:{pass}@localhost:{port}/postgres',
      }
    }

    creds = {'user': 'sysadmin', 'pass': 'password123'}

    parsed_db_urls = misc.parse_templated(templated_db_urls, extra_keywords=creds)

    expected_db_urls = {
      'dw_db_url': "postgresql://sysadmin:password123@localhost:5432/postgres"
    }

    self.assertEqual(parsed_db_urls, expected_db_urls)

    templated_schematable_urls = {
      'keywords': {
        'dw':'postgresql://{user}:{pass}@localhost:5432/dw'
      },
      'templates': {
        'product': '{dw}#product',
        'producer': '{dw}#producer',
      }
    }

    creds = {'user': 'sysadmin', 'pass': 'password123'}

    parsed_schematable_urls = misc.parse_templated(templated_schematable_urls, extra_keywords=creds)

    expected_schematable_urls = {
      'product': 'postgresql://sysadmin:password123@localhost:5432/dw#product',
      'producer': 'postgresql://sysadmin:password123@localhost:5432/dw#producer',
    }

    self.assertEqual(parsed_schematable_urls, expected_schematable_urls)

  def test__pen2dt(self):

    pen = pendulum.create(2021,1,11)
    dt = misc.pen2dt(pen)
    self.assertEqual(type(dt), datetime.datetime)


  def test__timeit(self):
    sleeper = lambda duration: time.sleep(duration)

    try:
      timed_sleeper = misc.timeit(sleeper)
      timed_sleeper(1)
    except:
      self.fail('timeit decorated method raised an expection')

  def test__sql_arr_lit(self):

      self.assertEqual(
          misc.sql_arr_lit(['hello', 'world']), 
          "('hello', 'world')"
      )

      self.assertEqual(
          misc.sql_arr_lit([1, 2]), 
          "(1, 2)"
      )

      self.assertEqual(
          misc.sql_arr_lit(['hello']), 
          "('hello')"
      )

      self.assertRaises(
          AssertionError,
          lambda: misc.sql_arr_lit([])
      )


unittest.main()


from setuptools import setup, find_packages

setup(
    name='utilita',
    version='1.0.14',
    packages=find_packages(),
    url = 'https://github.com/json2d/utilita',
    license= 'MIT', # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    author = 'Jason Yung, Tommy Rojo',
    author_email = 'json.yung@gmail.com, tr.trojo@gmail.com',
    description = 'a utility library', # Give a short description about your library
    long_description = open("README.md").read(),
    long_description_content_type='text/markdown',
    install_requires= ['isoweek', 'pendulum>=2', 'deprecated', 'openpyxl>=3,<4', 'pandas>=1', 'sendgrid>=6.0.0'],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License'
    ]
)